import numpy as np

from glue.core import HubListener
from specviz.third_party.glue.data_viewer import dispatch as specviz_dispatch

from ..messages import (SliceIndexUpdateMessage, WavelengthUpdateMessage,
                        WavelengthUnitUpdateMessage, RedshiftUpdateMessage)
from .wavelengths import REST_WAVELENGTH_TEXT, OBS_WAVELENGTH_TEXT

RED_BACKGROUND = "background-color: rgba(255, 0, 0, 128);"

import logging
logging.basicConfig(format='%(levelname)-6s: %(name)-10s %(asctime)-15s  %(message)s')
log = logging.getLogger("SliceController")
log.setLevel(logging.DEBUG)

class SliceController(HubListener):

    def __init__(self, cubeviz_layout):
        self._cv_layout = cubeviz_layout
        self._hub = cubeviz_layout.session.hub
        ui = cubeviz_layout.ui

        # These are the contents of the text boxes
        self._slice_textbox = ui.slice_textbox
        self._wavelength_textbox = ui.wavelength_textbox

        # This is the slider widget itself
        self._slice_slider = ui.slice_slider

        # This is the label for the wavelength units
        self._wavelength_textbox_label = ui.wavelength_textbox_label

        self._slice_slider.valueChanged.connect(self._on_slider_change)
        self._slice_slider.sliderPressed.connect(self._on_slider_pressed)
        self._slice_slider.sliderReleased.connect(self._on_slider_released)
        self._slider_flag = False

        self._slice_textbox.returnPressed.connect(self._on_text_slice_change)
        self._wavelength_textbox.returnPressed.connect(self._on_text_wavelength_change)

        self._slice_slider.setEnabled(False)
        self._slice_textbox.setEnabled(False)
        self._wavelength_textbox.setEnabled(False)
        self._wavelength_textbox_label.setWordWrap(True)

        # This should be used to distinguised between observed and rest wavelengths
        # We are not going to enforce what the name should be at this level.
        self._wavelength_label_text = OBS_WAVELENGTH_TEXT

        self._wavelength_format = '{:.3}'
        self._wavelength_units = None
        self._wavelengths = None

        # Tracks the index of the synced viewers
        self.synced_index = None

        # Connect this class to specviz's event dispatch so methods can listen
        # to specviz events
        specviz_dispatch.setup(self)

    def enable(self):
        """
        Setup the slice slider (min/max, units on description and initial position).

        :return:
        """
        self.set_enabled(True)

        self._hub.subscribe(self, SliceIndexUpdateMessage, handler=self._handle_index_update)
        self._hub.subscribe(self, WavelengthUpdateMessage, handler=self._handle_wavelength_update)
        self._hub.subscribe(self, WavelengthUnitUpdateMessage, handler=self._handle_wavelength_units_update)
        self._hub.subscribe(self, RedshiftUpdateMessage, handler=self._handle_redshift_update)

        self._slice_slider.setMinimum(0)

    def _handle_wavelength_units_update(self, message):

        # Store the wavelength units and format
        self._wavelength_units = message.units
        self._wavelength_textbox_label.setText('{} ({})'.format(
            self._wavelength_label_text, self._wavelength_units))

    def _handle_wavelength_update(self, message):

        # Grab the wavelengths so they can be displayed in the text box
        self._wavelengths = message.wavelengths
        self._slice_slider.setMaximum(len(self._wavelengths) - 1)

        if self.synced_index is None:
            # Set the initial display to the middle of the cube
            middle_index = len(self._wavelengths) // 2
            self._slice_slider.setValue(middle_index)
            self.synced_index = middle_index

        index = self._cv_layout._active_cube._widget.slice_index
        self._wavelength_textbox.setText(self._wavelength_format.format(self._wavelengths[index]))

    def _handle_redshift_update(self, message):

        self._wavelength_label_text = message.label

        self._wavelength_textbox_label.setText('{} ({})'.format(
            self._wavelength_label_text, self._wavelength_units))

    def set_enabled(self, value):
        self._slice_slider.setEnabled(value)
        self._slice_textbox.setEnabled(value)
        self._wavelength_textbox.setEnabled(value)

    def update_index(self, index):
        self._slice_slider.setValue(index)

    def change_slider_value(self, amount):
        new_index = self._slice_slider.value() + amount
        self._slice_slider.setValue(new_index)

        specviz_dispatch.changed_dispersion_position.emit(pos=new_index)

    def _handle_index_update(self, message):
        index = message.index

        try:
            tb_index = int(self._slice_textbox.text())
        except ValueError:
            tb_index = -1

        if tb_index != index:
            self._slice_textbox.setText(str(index))

        try:
            wavelength = float(self._wavelength_textbox.text())
            wv_index = np.argsort(abs(self._wavelengths - wavelength))[0]
        except ValueError:
            wavelength = -1
            wv_index = -1

        self._wavelength_textbox.setText(self._wavelength_format.format(self._wavelengths[index]))

        slider_index = self._slice_slider.value()
        if slider_index != index:
            self._slice_slider.setValue(index)

        if self._cv_layout._active_cube._widget.synced:
            self.synced_index = index

        specviz_dispatch.changed_dispersion_position.emit(pos=index)

    def _on_slider_change(self, event):
        """
        Callback for change in slider value.

        :param event:
        :return:
        """
        index = self._slice_slider.value()
        self._send_index_message(index)

    def _send_index_message(self, index):
        msg = SliceIndexUpdateMessage(self, index,
                                      self._cv_layout.session.data_collection[0],
                                      slider_down=self._slider_flag)
        self._hub.broadcast(msg)

    def _on_slider_pressed(self):
        """
        Callback for slider pressed.
        activates fast_draw_slice_at_index flags
        """
        # This flag will activate fast_draw_slice_at_index
        # Which will redraw sliced images quickly
        self._slider_flag = True

    @specviz_dispatch.register_listener("finished_position_change")
    def _on_slider_released(self):
        """
        Callback for slider released (includes specviz slider).
        Dactivates fast_draw_slice_at_index flags
        Will do a full redraw of all synced viewers.
        This is considered the final redraw after fast_draw_slice_at_index
        blits images to the viewers. This function will redraw the axis,
        tites, labels etc...
        """
        # This flag will deactivate fast_draw_slice_at_index
        self._slider_flag = False

        index = self._slice_slider.value()
        self._send_index_message(index)
        specviz_dispatch.changed_dispersion_position.emit(pos=index)

    #@glue_subscribe(SliceIndexUpdateMessage)
    def _update_slice_textboxes(self, message):
        """
        Update the slice index number text box and the wavelength value displayed in the wavelengths text box.

        :param index: Slice index number displayed.
        :return:
        """
        index = message.index

        # Update the input text box for slice number
        self._slice_textbox.setText(str(index))

        # Update the wavelength for the corresponding slice number.
        self._wavelength_textbox.setText(self._wavelength_format.format(self._wavelengths[index]))

    def _on_text_slice_change(self, event=None):
        """
        Callback for a change in the slice index text box.  We will need to
        update the slider and the wavelength value when this changes.

        :param event:
        :return:
        """
        # Get the value they typed in, but if not a number, then let's just use
        # the first slice.
        try:
            index = int(self._slice_textbox.text())
            self._slice_textbox.setStyleSheet("")
        except ValueError:
            self._slice_textbox.setStyleSheet(RED_BACKGROUND)
            return

        if index == self._slice_textbox.text():
            return

        # If a number and out of range then set to the first or last slice
        # depending if they set the number too low or too high.
        if index < 0:
            index = 0
        if index > len(self._wavelengths) - 1:
            index = len(self._wavelengths) - 1

        self._send_index_message(index)

    def _on_text_wavelength_change(self, event=None, pos=None):
        """
        Callback for a change in wavelength input box. We want to find the
        closest wavelength and use the index of it.  We will need to update the
        slice index box and slider as well as the image.

        :param event:
        :param pos: This is the argument used by the specviz event listener to
                    update the CubeViz slider and associated text based on the
                    movement of the SpecViz position bar. The name of this
                    argument cannot change since it is the one expected by the
                    SpecViz event system.
        :return:
        """
        try:
            # Find the closest real wavelength and use the index of it
            wavelength = pos if pos is not None else float(self._wavelength_textbox.text())
            index = np.argsort(abs(self._wavelengths - wavelength))[0]
            self._wavelength_textbox.setStyleSheet("")
        except ValueError:
            self._wavelength_textbox.setStyleSheet(RED_BACKGROUND)
            return

        self._send_index_message(index)

    @specviz_dispatch.register_listener("change_dispersion_position")
    def specviz_wavelength_slider_change(self, event=None, pos=None):
        """
        SpecViz slider index changed callback
        """
        # if self._slider_flag is active then
        # something else is using it so don't
        # deactivate it when done (deactivate_flag)
        if self._slider_flag:
            deactivate_flag = False
        else:
            deactivate_flag = True
            self._slider_flag = True

        # The "pos" value coming from specviz appears to be related to the
        # index in the observed wavelength and so if there is a redshift
        # then we need to convert the pos to the rest wavelength position.
        if self._cv_layout._wavelength_controller and not self._cv_layout._wavelength_controller.redshift_z == 0.0:
            rest_wavelength = pos / (1 + self._cv_layout._wavelength_controller.redshift_z)
            pos = np.argsort(abs(self._wavelengths - rest_wavelength))[0]

            # Pos is a wavelength and not an index for the call back for specviz
            pos = self._wavelengths[pos]

        self._on_text_wavelength_change(event, pos)

        if deactivate_flag:
            self._slider_flag = False



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import os

import pytest
import numpy as np

from qtpy import QtCore
from glue.core import roi
from cubeviz.tools.collapse_cube import CollapseCube

from ...tests.helpers import (toggle_viewer, select_viewer, left_click,
                      left_button_press, right_button_press, enter_slice_text,
                      assert_all_viewer_indices, assert_slice_text)


DATA_LABELS = ['018.DATA', '018.NOISE']


@pytest.fixture(scope='module')
def collapse_cube(cubeviz_layout):
    cl = cubeviz_layout

    wavelengths = cubeviz_layout._wavelength_controller.wavelengths
    wavelength_units = cubeviz_layout._wavelength_controller._current_units

    cc = CollapseCube(wavelengths, wavelength_units, cl._data, parent=cl)

    return cc


def assert_red_stylesheet(widget):
    assert widget.styleSheet() == "color: rgba(255, 0, 0, 128)"


def test_collapse_ui(qtbot, collapse_cube):

    cc = collapse_cube
    cc.ui.start_input.setText('a')
    qtbot.mouseClick(cc.ui.calculate_button, QtCore.Qt.LeftButton)
    assert_red_stylesheet(cc.ui.start_label)

    cc.ui.start_input.setText('100')
    cc.ui.end_input.setText('50')
    qtbot.mouseClick(cc.ui.calculate_button, QtCore.Qt.LeftButton)
    assert_red_stylesheet(cc.ui.start_label)

    # "reset" the values
    cc.ui.start_input.setText('0')
    cc.ui.end_input.setText('1')

    # Negative simple sigma value
    cc.ui.sigma_combobox.setCurrentIndex(1)
    cc.ui.simple_sigma_input.setText('-10')
    qtbot.mouseClick(cc.ui.calculate_button, QtCore.Qt.LeftButton)
    assert_red_stylesheet(cc.ui.simple_sigma_label)

    # Negative advanced sigma value
    cc.ui.sigma_combobox.setCurrentIndex(2)
    cc.ui.advanced_sigma_input.setText('-10')
    qtbot.mouseClick(cc.ui.calculate_button, QtCore.Qt.LeftButton)
    assert_red_stylesheet(cc.ui.advanced_sigma_label)

    # Negative advanced sigma value
    cc.ui.sigma_combobox.setCurrentIndex(2)
    cc.ui.advanced_sigma_input.setText('3')
    cc.ui.advanced_sigma_lower_input.setText('2')
    cc.ui.advanced_sigma_upper_input.setText('1')
    qtbot.mouseClick(cc.ui.calculate_button, QtCore.Qt.LeftButton)
    assert_red_stylesheet(cc.ui.advanced_sigma_lower_label)
    assert_red_stylesheet(cc.ui.advanced_sigma_upper_label)

def test_starting_state(cubeviz_layout):

    cc = collapse_cube(cubeviz_layout)

    # No clipping
    data_name = DATA_LABELS[0]
    operation = 'Sum'
    spatial_region = 'Image'
    start_index = 100
    end_index = 300
    sigma_selection = 'No Sigma Clipping'
    sigma_parameter = None

    new_wavelengths, new_component, label = cc._calculate_collapse(data_name,
            operation, spatial_region, sigma_selection, sigma_parameter, start_index, end_index)

    assert label == '018.DATA-collapse-Sum (1.9531e-06, 2.0092e-06)'

    output = np.array([[0, 0, 0, 0], [139.281, 1088, 844.167, 739.379]])
    assert np.allclose(new_component[:2,:4], output, atol=1)


    # Simple Sigma Clipping
    data_name = DATA_LABELS[0]
    operation = 'Sum'
    spatial_region = 'Image'
    start_index = 300
    end_index = 400
    sigma_selection = 'Simple Sigma Clipping'
    sigma_parameter = 3.0

    new_wavelengths, new_component, label = cc._calculate_collapse(data_name,
            operation, spatial_region, sigma_selection, sigma_parameter, start_index, end_index)

    assert label == '018.DATA-collapse-Sum (2.0092e-06, 2.0373e-06) sigma=3.0'

    output = np.array([[0, 0, 0, 0], [845.868, 886.972, 726.566, 976.271]])
    assert np.allclose(new_component[:2,:4], output, atol=1)

    # Advanced Sigma Clipping
    data_name = DATA_LABELS[0]
    operation = 'Sum'
    spatial_region = 'Image'
    start_index = 300
    end_index = 400
    sigma_selection = 'Advanced Sigma Clipping'
    sigma_parameter = [3.0, 1.0, 4.0, 2]

    new_wavelengths, new_component, label = cc._calculate_collapse(data_name,
            operation, spatial_region, sigma_selection, sigma_parameter, start_index, end_index)

    assert label == '018.DATA-collapse-Sum (2.0092e-06, 2.0373e-06) sigma=3.0 sigma_lower=1.0 sigma_upper=4.0 sigma_iters=2'

    output = np.array([[0, 0, 0, 0], [867.687, 956.132, 775.984, 1008]])
    assert np.allclose(new_component[:2,:4], output, atol=1)

def test_regions(qtbot, cubeviz_layout):

    viewer = cubeviz_layout.split_views[0]._widget
    # Create a pretty arbitrary circular ROI
    viewer.apply_roi(roi.CircularROI(xc=6, yc=10, radius=3))

    cc = collapse_cube(cubeviz_layout)

    start_index = 682
    end_index = 1364

    # Set the values and do the calculation through the GUI
    cc.ui.data_combobox.setCurrentIndex(0)
    cc.ui.operation_combobox.setCurrentIndex(0)
    cc.ui.spatial_region_combobox.setCurrentIndex(1)
    cc.ui.region_combobox.setCurrentIndex(1) # indices
    cc.ui.start_input.setText('{}'.format(start_index))
    cc.ui.end_input.setText('{}'.format(end_index))
    cc.ui.sigma_combobox.setCurrentIndex(0)
    qtbot.mouseClick(cc.ui.calculate_button, QtCore.Qt.LeftButton)

    # Calculate what we expect
    np_data = cubeviz_layout._data[DATA_LABELS[0]]
    mask = cubeviz_layout._data.subsets[0].to_mask()
    np_data_sum = np.sum(np_data[start_index:end_index]*mask[start_index:end_index], axis=0)

    # Get the result
    collapse_component_id = [str(x) for x in cubeviz_layout._data.container_2d.component_ids() if str(x).startswith('018.DATA-collap')][0]
    np_result = cubeviz_layout._data.container_2d[collapse_component_id]

    # Delete the ROI first, in case the assert fails
    dc = cubeviz_layout.session.application.data_collection
    dc.remove_subset_group(dc.subset_groups[0])

    print('combo box items {}'.format([cc.ui.operation_combobox.itemText(i) for i in range(cc.ui.operation_combobox.count())]))
    print('combo box selected {}'.format(cc.ui.operation_combobox.currentText()))
    print('np_data_sum {}'.format(np_data_sum))
    print('np_result {}'.format(np_result))
    assert np.allclose(np_data_sum, np_result, atol=1.0)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest
from qtpy import QtCore

import numpy as np

from ...tests.helpers import (enter_slice_text, enter_wavelength_text,
                              left_click, select_viewer, enter_slice_text,
                              toggle_viewer, assert_viewer_indices,
                              assert_all_viewer_indices,
                              assert_wavelength_text, assert_slice_text)


def set_slider_index(layout, index):
    layout._slice_controller._slice_slider.setSliderPosition(index)

def find_nearest_slice(wavelengths, value):
    return np.argsort(abs(wavelengths - value))[0]

def all_but_index(array, index):
    return array[:index] + array[index+1:]

@pytest.fixture(scope='module')
def cube_bounds(cubeviz_layout):
    wavelengths = cubeviz_layout._wavelength_controller.wavelengths

    bounds = {
        'max_slice': len(cubeviz_layout._data['018.DATA']) - 1,
        'wavelengths': wavelengths,
        'min_wavelength': wavelengths[0],
        'max_wavelength': wavelengths[-1]
    }

    yield bounds

@pytest.mark.parametrize('slice_index', [0, 100, 1000, 1024])
def test_enter_valid_slice_text(qtbot, cubeviz_layout, slice_index):
    enter_slice_text(qtbot, cubeviz_layout, slice_index)
    assert_all_viewer_indices(cubeviz_layout, slice_index)
    assert_slice_text(cubeviz_layout, slice_index)

def test_enter_oob_slice_text(qtbot, cubeviz_layout, cube_bounds):
    # Enter a negative slice value
    enter_slice_text(qtbot, cubeviz_layout, "-42")
    assert_all_viewer_indices(cubeviz_layout, 0)
    assert_slice_text(cubeviz_layout, 0)

    max_slice = cube_bounds['max_slice']

    # Enter an impossibly large slice value
    enter_slice_text(qtbot, cubeviz_layout, str(2**20))
    assert_all_viewer_indices(cubeviz_layout, max_slice)
    assert_slice_text(cubeviz_layout, max_slice)

@pytest.mark.parametrize('bad_text', ['garbage', '1e-07', '3.14', ''])
def test_garbage_slice_text(qtbot, cubeviz_layout, bad_text):
    # Get the current index of all viewers (since all are currently synced)
    slice_index = cubeviz_layout.synced_index

    # Make sure that entering garbage text does not change the index
    enter_slice_text(qtbot, cubeviz_layout, bad_text)
    assert_all_viewer_indices(cubeviz_layout, slice_index)
    assert_slice_text(cubeviz_layout, bad_text)

@pytest.mark.parametrize('slice_index', [0, 100, 1000, 1024, 2000])
def test_wavelength_slider(cubeviz_layout, slice_index, cube_bounds):
    set_slider_index(cubeviz_layout, slice_index)
    assert_all_viewer_indices(cubeviz_layout, slice_index)

    # Make sure that wavelength text matches slice value
    wavelength_text = "{:.3}".format(cube_bounds['wavelengths'][slice_index])
    assert_wavelength_text(cubeviz_layout, wavelength_text)

# These wavelengths are tuned to test the data file data_cube.fits.gz
@pytest.mark.parametrize('wavelength',
    [1.8421e-06, 1.931947e-06, 2.0104081e-06, 2.122048e-06, 2.270e-06, 2.51149e-06])
def test_nearest_slice_index(qtbot, cubeviz_layout, wavelength, cube_bounds):
    # While this test mostly just duplicates the code in
    # _on_text_wavelength_change, it is useful to make sure that the effects
    # carry through in the viewer slices and text boxes
    wavelengths = cube_bounds['wavelengths']

    enter_wavelength_text(qtbot, cubeviz_layout, str(wavelength))
    slice_index = find_nearest_slice(wavelengths, wavelength)
    assert_all_viewer_indices(cubeviz_layout, slice_index)
    assert_slice_text(cubeviz_layout, slice_index)
    wavelength_text = "{:.3}".format(wavelengths[slice_index])
    assert_wavelength_text(cubeviz_layout, wavelength_text)

def test_enter_oob_wavelength_text(qtbot, cubeviz_layout, cube_bounds):
    enter_wavelength_text(qtbot, cubeviz_layout, '0')
    assert_all_viewer_indices(cubeviz_layout, 0)
    # We use the slice value as a proxy for the correct wavelength value
    assert_slice_text(cubeviz_layout, 0)

    max_slice = cube_bounds['max_slice']

    # enter an impossibly large wavelength value
    enter_wavelength_text(qtbot, cubeviz_layout, '42')
    assert_all_viewer_indices(cubeviz_layout, max_slice)
    assert_slice_text(cubeviz_layout, max_slice)

    # enter a negative wavelength value
    enter_wavelength_text(qtbot, cubeviz_layout, '-42')
    assert_all_viewer_indices(cubeviz_layout, 0)
    assert_slice_text(cubeviz_layout, 0)

@pytest.mark.parametrize('bad_text', ['garbage', ''])
def test_garbage_wavelength_text(qtbot, cubeviz_layout, bad_text):
    # Get the current index of all viewers (since all are currently synced)
    slice_index = cubeviz_layout.synced_index

    # Make sure that entering garbage text does not change the index
    enter_wavelength_text(qtbot, cubeviz_layout, bad_text)
    assert_all_viewer_indices(cubeviz_layout, slice_index)
    assert_wavelength_text(cubeviz_layout, bad_text)

def unsync_and_verify(qtbot, layout, sync_params, synced_index, unsynced_index):
    unsynced_viewer = sync_params['unsynced_viewer']

    # Unsync the selected viewer
    left_click(qtbot, sync_params['checkbox'])
    # At this point no slice indices should have changed yet
    assert layout.synced_index == synced_index
    assert_all_viewer_indices(layout, synced_index)

    # Activate the unsynced viewer
    select_viewer(qtbot, unsynced_viewer)
    # Change the slice index
    enter_slice_text(qtbot, layout, str(unsynced_index))
    # Make sure the active (unsynced) viewer's index has been updated
    assert unsynced_viewer._widget.slice_index == unsynced_index
    # Make sure no other viewer indices have been updated
    assert_viewer_indices(sync_params['other_viewers'], synced_index)
    # Make sure the displayed slice index reflects the active viewer
    assert_slice_text(layout, str(unsynced_index))

def update_synced_index_and_verify(qtbot, layout, sync_params, new_synced_index,
                                   unsynced_index):
    old_synced_index = layout.synced_index
    unsynced_viewer = sync_params['unsynced_viewer']
    other_viewers = sync_params['other_viewers']

    select_viewer(qtbot, sync_params['synced_viewer'])
    assert_slice_text(layout, str(old_synced_index))

    # Make sure that no indices have changed
    assert unsynced_viewer._widget.slice_index == unsynced_index
    assert_viewer_indices(other_viewers, old_synced_index)

    # Update the synced index
    enter_slice_text(qtbot, layout, str(new_synced_index))
    assert unsynced_viewer._widget.slice_index == unsynced_index
    assert_viewer_indices(other_viewers, new_synced_index)
    import time
    time.sleep(0.5)
    assert_slice_text(layout, str(new_synced_index))

@pytest.fixture(scope='function')
def sync_params(cubeviz_layout, request):
    """This fixture expects indirect parametrization providing the viewer index"""
    viewer_index = request.param
    # See comment in ``test_unsync_viewer`` below about this workaround
    if viewer_index == 0:
        return None

    params = dict(
        checkbox=cubeviz_layout._synced_checkboxes[viewer_index],
        unsynced_viewer=cubeviz_layout.cube_views[viewer_index],
        other_viewers=all_but_index(cubeviz_layout.cube_views, viewer_index),
        synced_viewer=all_but_index(
            cubeviz_layout.split_views, viewer_index-1)[0]
    )

    return params

# The 0 index is used to enable a workaround for running the tests on Travis
# CI. The single cube viewer will be tested by a separate test case.
@pytest.mark.parametrize('sync_params', [0, 1, 2, 3], indirect=True)
def test_unsync_viewer(qtbot, cubeviz_layout, sync_params):
    # This is a workaround for an issue when running tests on Linux VMs in
    # Travis CI. It does not appear to be an actual bug with CubeViz, but it
    # might be due to a race condition between the application and the test
    # framework. The solution is to make sure to toggle the viewer mode prior
    # to running the rest of these tests. It will be toggled back to single
    # mode by the reset_state fixture defined in the top-level conftest.py.
    if sync_params is None:
        toggle_viewer(qtbot, cubeviz_layout)
        pytest.skip()

    # Get the current index of all viewers
    synced_index = cubeviz_layout.synced_index
    # Sanity check to make sure all viewers are actually synced
    assert_all_viewer_indices(cubeviz_layout, synced_index)

    unsynced_viewer = sync_params['unsynced_viewer']
    other_viewers = sync_params['other_viewers']

    # Activate the unsynced viewer
    unsynced_index = 42
    unsync_and_verify(
        qtbot, cubeviz_layout, sync_params, synced_index, unsynced_index)

    # Toggle to single image mode
    toggle_viewer(qtbot, cubeviz_layout)
    assert_slice_text(cubeviz_layout, str(synced_index))

    # Toggle back to split image mode
    toggle_viewer(qtbot, cubeviz_layout)
    assert_slice_text(cubeviz_layout, str(unsynced_index))

    synced_index = 1234
    # Now activate one of the synced viewers
    update_synced_index_and_verify(
        qtbot, cubeviz_layout, sync_params, synced_index, unsynced_index)

    # Activate the unsynced viewer again
    select_viewer(qtbot, unsynced_viewer)
    assert_slice_text(cubeviz_layout, str(unsynced_index))

    # Sync the unsynced viewer
    left_click(qtbot, sync_params['checkbox'])
    assert_all_viewer_indices(cubeviz_layout, synced_index)
    assert_slice_text(cubeviz_layout, str(synced_index))

@pytest.mark.parametrize('sync_params', [1, 2, 3], indirect=True)
def test_resync_inactive_viewer(qtbot, cubeviz_layout, sync_params):
    """In contrast to the previous test, this test makes sure the unsynced
    viewer is successfully re-synced even when it is not the active viewer.
    """
    # Get the current index of all viewers
    synced_index = cubeviz_layout.synced_index

    unsynced_viewer = sync_params['unsynced_viewer']
    other_viewers = sync_params['other_viewers']

    unsynced_index = 42
    unsync_and_verify(
        qtbot, cubeviz_layout, sync_params, synced_index, unsynced_index)

    synced_index = 1234
    # Now activate one of the synced viewers
    update_synced_index_and_verify(
        qtbot, cubeviz_layout, sync_params, synced_index, unsynced_index)

    # Sync the (inactive) unsynced viewer
    left_click(qtbot, sync_params['checkbox'])
    assert_all_viewer_indices(cubeviz_layout, synced_index)
    assert_slice_text(cubeviz_layout, str(synced_index))

    # Activate the unsynced viewer again
    select_viewer(qtbot, unsynced_viewer)
    assert_slice_text(cubeviz_layout, str(synced_index))

def test_multiple_unsynced_viewers(qtbot, cubeviz_layout):
    # Get the current index of all viewers
    synced_index = cubeviz_layout.synced_index
    # Sanity check to make sure all viewers are actually synced
    assert_all_viewer_indices(cubeviz_layout, synced_index)

    unsync1 = cubeviz_layout.cube_views[1]
    checkbox1 = cubeviz_layout._synced_checkboxes[1]
    unsync_index1 = 42

    unsync2 = cubeviz_layout.cube_views[3]
    checkbox2 = cubeviz_layout._synced_checkboxes[3]
    unsync_index2 = 1234

    remain_synced = [
        cubeviz_layout.cube_views[0],
        cubeviz_layout.cube_views[2]
    ]

    # Unsync the first viewer
    left_click(qtbot, checkbox1)
    # Activate the first viewer
    select_viewer(qtbot, unsync1)
    enter_slice_text(qtbot, cubeviz_layout, unsync_index1)
    assert_slice_text(cubeviz_layout, str(unsync_index1))
    assert_viewer_indices(remain_synced + [unsync2], synced_index)

    # Unsync the second viewer: the order of activation/unsync shouldn't
    # matter, so we do it differently here than we did above
    select_viewer(qtbot, unsync2)
    left_click(qtbot, checkbox2)
    # Nothing should have changed yet
    assert_slice_text(cubeviz_layout, str(synced_index))
    assert_viewer_indices(remain_synced + [unsync2], synced_index)
    assert_viewer_indices([unsync1], unsync_index1)

    # Update the second synced viewer
    enter_slice_text(qtbot, cubeviz_layout, unsync_index2)
    assert_slice_text(cubeviz_layout, str(unsync_index2))
    # Make sure the first unsynced viewer is unchanged
    assert_viewer_indices([unsync1], unsync_index1)
    # Make sure the synced viewers are unchanged
    assert_viewer_indices(remain_synced, synced_index)

    # Update the synced viewers
    synced_index = 1234
    select_viewer(qtbot, remain_synced[1])
    enter_slice_text(qtbot, cubeviz_layout, str(synced_index))
    assert_slice_text(cubeviz_layout, str(synced_index))
    assert_viewer_indices(remain_synced, synced_index)
    assert_viewer_indices([unsync1], unsync_index1)
    assert_viewer_indices([unsync2], unsync_index2)

    # Activate the first unsynced viewer again
    select_viewer(qtbot, unsync1)
    assert_slice_text(cubeviz_layout, str(unsync_index1))
    assert_viewer_indices([unsync2], unsync_index2)
    assert_viewer_indices(remain_synced, synced_index)

    # Sync the first unsynced viewer
    left_click(qtbot, checkbox1)
    assert_slice_text(cubeviz_layout, str(synced_index))
    assert_viewer_indices(remain_synced + [unsync1], synced_index)
    assert_viewer_indices([unsync2], unsync_index2)

    # Activate the second (and still) unsynced viewer again
    select_viewer(qtbot, unsync2)
    assert_slice_text(cubeviz_layout, str(unsync_index2))

    # Sync the remaining unsynced viewer
    left_click(qtbot, checkbox2)
    assert_slice_text(cubeviz_layout, str(synced_index))
    assert_all_viewer_indices(cubeviz_layout, synced_index)

# Wavelength redshift and units testing
def test_wavelength_ui(qtbot, cubeviz_layout, slice_index=1000):
    from ...tools.wavelengths_ui import WavelengthUI

    wui = WavelengthUI(cubeviz_layout._wavelength_controller)

    # First move to the a specific slice
    enter_slice_text(qtbot, cubeviz_layout, slice_index)
    assert_slice_text(cubeviz_layout, slice_index)
    assert_wavelength_text(cubeviz_layout, '2.21e-06')

    # Now change the units
    wui.do_calculation(wavelength_redshift=0, wavelength_units='um')
    assert_wavelength_text(cubeviz_layout, '2.21')

    # Units and redshift
    wui.do_calculation(wavelength_redshift=1, wavelength_units='um')
    assert_wavelength_text(cubeviz_layout, '1.1')

    # Units and negative redshift
    wui.do_calculation(wavelength_redshift=-3, wavelength_units='um')
    assert_wavelength_text(cubeviz_layout, '-1.1')

    # with the units and redshift, go to another slice
    enter_slice_text(qtbot, cubeviz_layout, 2000)
    assert_wavelength_text(cubeviz_layout, '-1.24')

    wui.do_calculation(wavelength_redshift=0, wavelength_units='m')
